import numpy as np
import pandas as pd


def create_random_normal_data(nb_rows, nb_numerical_col, mu, sigma, nb_categorical_col, timeseries=False):
    np.random.seed(42)
    cat_value = 'a'
    data = {}
    for i in range(nb_numerical_col):
        col_name = 'col_num' + str(i)
        data[col_name] = np.random.normal(mu, sigma, size=nb_rows)
    for j in range(nb_categorical_col):
        col_name = 'col_cat' + str(j)
        data[col_name] = [cat_value] * nb_rows
    df = pd.DataFrame(data)
    if timeseries:
        df = _transform_to_timeseries_data(df)
    return df


def create_simple_data(nb_rows, nb_numerical_col, nb_categorical_col, timeseries=False):
    np.random.seed(42)
    cat_value = 'a'
    data = {}
    for i in range(nb_numerical_col):
        col_name = 'col_num' + str(i)
        data[col_name] = range(0, nb_rows)
    for j in range(nb_categorical_col):
        col_name = 'col_cat' + str(j)
        data[col_name] = [cat_value] * nb_rows
    df = pd.DataFrame(data)
    if timeseries:
        df = _transform_to_timeseries_data(df)
    return df


def _transform_to_timeseries_data(df):
    start = pd.to_datetime('2018-08-08')
    rng = pd.date_range(start, periods=len(df))
    df['DateTime'] = rng

    ts = df.set_index('DateTime')
    ts.index = pd.to_datetime(ts.index)
    return ts



# -*- coding: utf-8 -*-

import unittest

import pandas as pd

from mdspy.data_creation import create_random_normal_data


class Spark_test(unittest.TestCase):
    """Basic test cases."""

    def test_random_data(self):
        # Testing case with both kinds of columns
        df = create_random_normal_data(10, 3, 5, 6, 4)
        expected_len = 10
        actual_len = len(df)
        expected_width = 7
        actual_width = len(df.columns)
        self.assertEqual(expected_len, actual_len)
        self.assertEqual(expected_width, actual_width)

        # Testing case with no num columns
        df2 = create_random_normal_data(10, 0, 5, 6, 4)
        expected_len = 10
        actual_len = len(df2)
        expected_width = 4
        actual_width = len(df2.columns)

        self.assertEqual(expected_len, actual_len)
        self.assertEqual(expected_width, actual_width)

        # Testing case no cat columns

        df2 = create_random_normal_data(10, 3, 5, 6, 0)
        expected_len = 10
        actual_len = len(df2)
        expected_width = 3
        actual_width = len(df2.columns)

        self.assertEqual(expected_len, actual_len)
        self.assertEqual(expected_width, actual_width)

    def test_random_timeseries_data(self):
        # Testing case with both kinds of columns
        df = create_random_normal_data(10, 3, 5, 6, 4, timeseries=True)
        expected_datetime_value_type = type(pd.to_datetime('2018-01-01'))
        actual_datetime_value_type = type(df.index[0])

        self.assertEqual(expected_datetime_value_type, actual_datetime_value_type)



import pandas as pd
from matplotlib import pyplot as plt
from sklearn.decomposition import PCA


def create_sliding_windows(df, col_name, window_size):
    res = df.copy()
    for i in range(1, window_size + 1):
        res[col_name + '_' + str(i)] = res[col_name].shift(i)
        cols = [col_name] + [col_name + '_' + str(i) for i in range(1, window_size + 1)]
        cols.reverse()
    res = res[cols].dropna(axis=0)
    return res


class ts_pca:
    def __init__(self, df, n_components=3):
        self.pca = PCA(n_components=n_components)
        self.df = df
        self.pca.fit(df)
        self._projected = None

    def get_PCA(self):
        return self.pca

    def transform(self, df=None):
        col_names = ['PCA_' + str(i) for i in range(1, self.pca.n_components_ + 1)]
        if df is None:
            if self._projected is not None:
                return self._projected
            self._projected = pd.DataFrame(self.pca.transform(self.df), index=self.df.index, columns=col_names)

            return self._projected
        res = pd.DataFrame(self.pca.transform(df), index=df.index, columns=col_names)
        return res

    def get_report(self):
        return 'number of components: ' + str(self.pca.n_components) + '\n' + \
               'explained variance: ' + str(self.pca.explained_variance_ratio_)

    def plot_coefficients(self):
        fig, axs = plt.subplots(self.pca.n_components, 1, sharex=True)
        for i in range(self.pca.n_components):
            axs[i].bar(range(len(self.pca.components_[i])), self.pca.components_[i], label='PCA_' + str(i + 1))
        fig.subplots_adjust(hspace=0)
        plt.show()

    def inverse_transform(self):
        if self._projected is None:
            self.transform()
        return pd.DataFrame(self.pca.inverse_transform(self._projected), index=self._projected.index,
                            columns=self.df.columns)



import numpy as np
import pandas as pd

from mdspy.PCA_timeseries import ts_pca, create_sliding_windows


def feature_creators(list_functions, history_size, big_data=False):
    def feats_and_concat(small_df):
        small_df.index = small_df.index.get_level_values(0)
        index = small_df.index[-1]
        line = pd.DataFrame()
        for func in list_functions:
            tmp = func(small_df)
            line = pd.concat([line, tmp], axis=1)
        line.index = [index]
        feature_creators.res = pd.concat([feature_creators.res, line], axis=0)

    def run(df):
        feature_creators.res = pd.DataFrame()
        # df = df.reset_index(drop=True)

        if big_data:
            for i in range(len(df.index) - history_size + 1):
                sub_df = df[i:i + history_size]
                feats_and_concat(sub_df)
        else:
            _roll(df, history_size=history_size).apply(feats_and_concat)
        return feature_creators.res

    return run


def fe_get_day_of_week(name_of_feature):
    def get_day_of_week(df_small):
        return pd.DataFrame({name_of_feature: [df_small.index[-1].weekday()]})

    return get_day_of_week


def fe_get_lag(name_of_feature, target, lag=1):
    def get_lag(df_small):
        return pd.DataFrame({name_of_feature: [df_small[target].shift(lag).values[-1]]})

    return get_lag


def fe_get_diff(name_of_feature, target, jump=2):
    def get_diff(df_small):
        return pd.DataFrame({name_of_feature: [df_small[target].diff(jump).values[-1]]})

    return get_diff


def fe_get_pca(name_of_feature, target, nb_component, df, windows_size):
    df = create_sliding_windows(df, target, windows_size)
    pca = ts_pca(df.reset_index(drop=True), n_components=nb_component)

    def get_pca(df_small):
        df_small_windows = create_sliding_windows(df_small, target, windows_size)
        df_with_pca = pca.transform(df_small_windows)
        data_dict = {}
        for i in range(nb_component):
            data_dict[name_of_feature + '_' + str(i + 1)] = [df_with_pca['PCA' + '_' + str(i + 1)].values[-1]]

        res = pd.DataFrame(data_dict)

        return res

    return get_pca


def _roll(df, history_size, **kwargs):
    df = df.dropna(axis=0)
    roll_array = np.dstack([df.values[i:i + history_size, :] for i in range(len(df.index) - history_size + 1)]).T
    panel = pd.Panel(roll_array,
                     items=df.index[history_size - 1:],
                     major_axis=df.columns, minor_axis=pd.Index(range(history_size), name='roll'))
    df_group = panel.to_frame().unstack().T.groupby(level=0, **kwargs)
    return df_group



# -*- coding: utf-8 -*-

import unittest

from mdspy.data_creation import create_random_normal_data
from mdspy.spark import run_spark_job_concat_dfs, run_spark_job_no_reduce, init_spark


class Spark_test(unittest.TestCase):
    """Basic test cases."""

    @classmethod
    def setUpClass(cls):
        print('i am here')
        cls.sc = init_spark()

    def test_spark_concat_dfs(self):
        # df = pd.read_csv('../data/energydata_complete.csv')
        df = create_random_normal_data(10, 2, 3, 4, 5)
        parameters = [df.head(10), df.tail(10)]
        parameters_devices = [['dev1', 'dev2', 'dev3'], ['dev4', 'dev5', 'dev6']]

        def funct(df):
            return df.head(5)

        df_res = run_spark_job_concat_dfs(parameters, funct, nb_cores='*', sc=Spark_test.sc)
        expected = 10
        actual = len(df_res)
        self.assertEqual(expected, actual)

    def test_spark_no_reduce_dfs(self):
        # df = pd.read_csv('../data/energydata_complete.csv')
        df = create_random_normal_data(10, 2, 3, 4, 5)
        parameters = [df.head(10), df.tail(10)]
        parameters_devices = [['dev1', 'dev2', 'dev3'], ['dev4', 'dev5', 'dev6']]

        def funct(df):
            return df.head(5)

        run_spark_job_no_reduce(parameters, funct, nb_cores='*', sc=Spark_test.sc)

        self.assertTrue(True)

    @classmethod
    def tearDownClass(cls):
        cls.sc.stop()



# -*- coding: utf-8 -*-

import unittest

from mdspy.data_creation import create_simple_data
from mdspy.features import fe_get_day_of_week, feature_creators, fe_get_lag, fe_get_diff


class Features_test(unittest.TestCase):
    """Basic test cases."""

    def test_feature_day_of_week(self):
        df = create_simple_data(10, 50, 0, timeseries=True)
        func_dow = fe_get_day_of_week('dow')
        run = feature_creators([func_dow], history_size=2)
        res = run(df)
        print(res)
        expected_shape = (9, 1)
        actual_shape = res.shape
        self.assertEqual(expected_shape, actual_shape)
        self.assertEqual(28, res['dow'].sum())

    def test_lag(self):
        df = create_simple_data(10, 50, 0, timeseries=True)
        print('here df', df)
        func_dow = fe_get_lag('col_num1_lag_1', 'col_num1', lag=1)
        run = feature_creators([func_dow], history_size=4)
        res = run(df)
        print('here res', res)
        res.to_csv('check_this_out.csv')
        expected_shape = (7, 1)
        actual_shape = res.shape
        self.assertEqual(expected_shape, actual_shape)
        self.assertEqual(35, res['col_num1_lag_1'].sum())

    def test_diff(self):
        df = create_simple_data(10, 50, 0, timeseries=True)
        print('here df', df)
        func_dow = fe_get_diff('col_num1_lag_1', 'col_num1', jump=2)
        run = feature_creators([func_dow], history_size=4)
        res = run(df)
        print('here res', res)
        res.to_csv('check_this_out.csv')
        expected_shape = (7, 1)
        actual_shape = res.shape
        self.assertEqual(expected_shape, actual_shape)
        self.assertEqual(14, res['col_num1_lag_1'].sum())
